/* 	DFT.C	/ Borlan C/C++ 3.1 /

	Sadrzi rutine za racunanje DFT/FFT digitalnog signala.

	Pavlovic Dejan, 1998.
*/

#include <math.h>
#include "dft.h"

void set_float_array(float far a[], int N, float value)
/* U niz upisuje vrednost value. */
{
   int i;

   for (i = 0; i < N; i++)
      a[i] = value;
}

void scale_float_array(float far a[], int N, float scale)
/* Skalira niz duzine N */
{
   int i;

   for (i = 0; i < N; i++)
      a[i] *= scale;
}

double I0(double x)
/* Baselova funkcija nultog reda. */
{
   int k;
   double x2 = x/2, xs=1.0, fact = 1.0, a = 1;
   double sum = 1.0;

   for (k = 1; fabs(a) > 1e-9; k++) {
      xs *= x2;
      fact *= k;

      a = xs/fact;
      sum += a*a;
   }

   return sum;
}

void make_window(float far w[], int N_win, int N, int scale,
		 enum window_type type,
		 double kaiser_beta,
		 float far w_load[], int N_w_load)
/* Kreira odgovarajucu prozorsku funkciju, duzine N_win i niz duzine N */
/* (N je duzina sempla). Ako je scale aktivno, vrsi se skaliranje tako */
/* da je sum(w[0:N_win-1]) = N*1 = N. Prom. window_type odrdjuje tip */
/* prozorske f-je. Ako je kaizerova onda se koristi param. kaiser_beta */
/* Ako je tipa LOAD onda se u w[0:N_win-1] upisuje w_load[0:N_w_load-1] */
/* Pri semu se vrsi linearna interpolacija ako je N_win <> N_w_load */
{
   int    n;
   double a0, a1, a2;
   double c, I0_beta;
   double pos, pos_scale;
   int    pos_int;
   double sum, c_scale;

   switch (type) {
      case WIN_NONE:
	 for (n = 0; n < N_win; n++)
	    w[n] = 1.0;
	 break;
      case WIN_BARTLETT:
	 for (n = (N_win + 1)/2-1; n >= 0; n--)
	    w[n] = 2.0*n/N_win;
	 for (n = N_win/2; n < N_win; n++)
	    w[n] = w[N_win-1 - n];
	 break;
      case WIN_SINE:
	 for (n = 0; n < N_win; n++)
	    w[n] = sin(n*M_PI/N_win);
	 break;
      case WIN_HANU:
	 for (n = 0; n < N_win; n++) {
	    w[n] = sin(n*M_PI/N_win);
	    w[n] *= w[n];
	 }
	 break;
      case WIN_HAMMING:
	 a0 = 25.0/46.0;
	 a1 = 1.0 - a0;
	 for (n = 0; n < N_win; n++)
	    w[n] = a0 - a1*cos(n*(2*M_PI)/N_win);
	 break;
      case WIN_BLACKMAN:
	 a0 = 0.42659071;
	 a1 = 0.49656062;
	 a2 = 0.07684867;
	 for (n = 0; n < N_win; n++)
	    w[n] = a0 - a1*cos(n*(2*M_PI)/N_win) + a2*cos(n*(4*M_PI)/N_win);
	 break;
      case WIN_KAISER:
	 I0_beta = I0(kaiser_beta);
	 for (n = 0; n < N_win; n++) {
	    c = 1.0 - 2.0*n/(double)(N_win - 1);
	    w[n] = I0(kaiser_beta*sqrt(1.0 - c*c)) / I0_beta;
	 }
	 break;
      case WIN_LOAD:
	 pos_scale = (double)(N_w_load - 1) / (double)(N_win - 1);
	 w[N_win - 1] = w_load[N_w_load - 1];
	 for (n = N_win - 2; n >= 0; n--) {
	    pos = pos_scale * n;
	    pos_int = pos;

	    w[n] = w_load[pos_int] +
		   (w_load[pos_int + 1] - w_load[pos_int]) * (pos - pos_int);
	 }
	 break;
   }

   if (scale) {			/* Skaliranje tako da je sum(w[]) = N */
      sum = 0.0;
      for (n = 0; n < N_win; n++)
	 sum += w[n];

      c_scale = N/sum;

      for (n = 0; n < N_win; n++)
	 w[n] *= c_scale;
   }

   for (n = N_win; n < N; n++)
      w[n] = 0.0;
}

void make_binindex(int far binindex[], int N, int dft_type)
/* Kreira niz binindex[] koji sadrzi informaciju kako se prerspodeljuju */
/* semplovi u vremenskom domenu u zavsnosti od tipa DFT. */
{
   int i, j;
   int it, jt, jtstart = N >> 1;

   if (dft_type == 0)
      for (i = 0; i < N; i++)
	 binindex[i] = i;
   else
      for (i = 0; i < N; i++) {
	 j = 0;
	 for (it = 1, jt = jtstart; it < N; it <<= 1, jt >>= 1)
	    if ((i & it) != 0)
	       j |= jt;
	 binindex[i] = j;
      }
}

void make_cossin_table(float far cosfi[], float far sinfi[], int N)
/* kreira tabelu kosinusa i sinusa potrebnih za izracunavanje DFT/FFT. */
{
   double c = -2*M_PI/N;
   int i;

   N /= 2;
   for (i = 0; i < N; i++) {
      cosfi[i] = cos(c*i);
      sinfi[i] = sin(c*i);
   }
}

void sample2bincomplex(float sample[], int far binindex[], float far w[],
		       cplx far x[], int N)
/* Konvertuje sempl u komleksne brojeve. Koristi prozorske f-je */
/* i vrsi odgovarajuce (binarno) premestanje u vremenskom domenu. */
{
   int i, j;

   for (i = 0; i < N; i++) {
      x[j = binindex[i]].re = *sample++ * *w++;
      x[j].im = 0.0;
   }
}

void realsample2bincomplex(float sample[], int far binindex[], float far w[],
			   cplx far x[], int N)
/* Konvertuje sempl duzine N u komleksnu sekvencu duzine N/2 */
/* radi brzeg izracunavanja DFT/FFT. Posle DFT/FFT treba izvrsiti */
/* korekciju. Koristi prozorske f-je i vrsi odgovarajuce */
/* (binarno) premestanje u vremenskom domenu (premestanje za FFT). */
{
   int i, j, i2;
   int N_2 = N/2;

   for (i = 0; i < N_2; i++) {
      i2 = i << 1;  /* i2 = i * 2 - parni indeksi */
      x[j = binindex[i]].re = sample[i2] * w[i2];
      i2++;	    /* neparni indeks */
      x[j].im = sample[i2] * w[i2];
   }
}

void correct_complex(cplx far x[], int N_new, cplx far *xp,
		     float far *pcos, float far *psin)
/* Ulaz je x = DFT(s1[] + s2[]*j), gde su s1[] i s2[] parni i neparni */
/* clanovi realnog niza. x trba modifikovati tako da se dobije spektar */
/* orginalne sekvence. xp[] je pomocna memorija duzine 2*size(x). */
/* pcos i psin su pokaz. na tabele cos/sin koje se koriste za izrac. FFT */
{
   int N = N_new / 2;
   cplx far *x1 = xp;       /* radi ustede memorije koristi se pomocni niz */
   cplx far *x2 = xp + N;   /* xp[] koji je u ovom trenutku slobodan */

   cplx far *px1 = x1;
   cplx far *px2 = x2;
   cplx far *px_begin = x;
   cplx far *px_end   = x + N - 1;
   cplx far *px_first;
   cplx far *px_second;
   double w_re, w_im, x2_re, x2_im, a, b;
   double alfa0 = M_PI/N;
   double wadd_re = cos(alfa0);
   double wadd_im = sin(alfa0);
   double w0_re, w0_im;
   int i, k;

   /* Kreira kompleksne nizove x1[] i x2[] koji predstavljaju */
   /* x1 = DFT(n1), x2 = DFT(n2), gde su n1 i n2 realni nizovi. */
   /* Kao ulaz je kompleksni niz x[] = DFT(n1 + j*n2) */
   px1->re = px_begin->re;
   px1->im = 0.0;
   px2->re = px_begin->im;
   px2->im = 0.0;
   px1++; px2++;
   px_begin++;
   for (i = 1; i < N; i++) {
      px1->re = 0.5*(px_begin->re + px_end->re);
      px1->im = 0.5*(px_begin->im - px_end->im);

      px2->re = 0.5*(px_begin->im + px_end->im);
      px2->im = 0.5*(px_end->re - px_begin->re);

      px1++;      px2++;
      px_begin++; px_end--;
   }

   /* Na osnovi x1 i x2 kreira niz x po formuli: */
   /* x[k] = x1[k] + W(k,2N)*x2[k], k = 0,...,N-1 */
   /* x[k + N] = x1[k] - W(k,2N)*x2[k], k = 0,...,N-1 */

   px1 = x1;     px2 = x2;
   px_first = x; px_second= x + N;
   for (k = 0; k < N; k++) {
      /* x[k] = x[k + N] = x1[k]; */
      *px_first = *px_second = *px1++;

      /* w_re = cos(k*M_PI/N);   */
      /* w_im = -sin(k*M_PI/N);  */

      if ((k & 1) == 0) {
	 w_re = *pcos++;
	 w_im = *psin++;
      }
      else {
	 w0_re = w_re;
	 w0_im = w_im;
	 w_re = w0_re*wadd_re + w0_im*wadd_im;
	 w_im = w0_im*wadd_re - w0_re*wadd_im;
      }

      /* x[k] += w * x2[k] */
      px_first->re += (a = w_re*(x2_re = px2->re) - w_im*(x2_im = px2->im));
      px_first->im += (b = w_re*x2_im + w_im*x2_re);

      /* x[k + N] -= w * x2[k] */
      px_second->re -= a;
      px_second->im -= b;

      px2++;
      px_first++;
      px_second++;
   }
}

void DFT(cplx far x[], cplx far f[], int N)
/* Orginalna funkcija za racunanje DFT, bez ikakvih optimizacija. */
{
   int k, n;
   double sre, sim, a, b, fi, c = 2.0*M_PI/N;

   for (k = 0; k < N; k++) {
      sre = 0.0;
      sim = 0.0;
      for (n = 0; n < N; n++) {
	 fi = c*k*(double)n;
	 a = cos(fi);
	 b = -sin(fi);
	 sre += x[n].re*a - x[n].im*b;
	 sim += x[n].re*b + x[n].im*a;
      }
      f[k].re = sre;
      f[k].im = sim;
   }
}

void FFT1(cplx far x[], int N)
/* Klasicni FFT radiks-2 algoritam sa preuredjenjem u vremenu (DIT). */
{
   int visina_leptira, visina_leptira_2, br_delova, deo, leptir, a, b, s;
   float fi, c = -2*M_PI/N;
   float wre, wim;
   float xbre, xbim;

   for (visina_leptira = 1; visina_leptira < N; visina_leptira <<= 1) {
      br_delova = N / (visina_leptira_2 = visina_leptira << 1);
      for (leptir = 0, s = 0; leptir < visina_leptira; leptir++, s += br_delova) {
	 fi = c * s;
	 wre = cos(fi);
	 wim = sin(fi);
	 for (a = leptir; a < N; a += visina_leptira_2) {
	    b = a + visina_leptira;

	    xbre = x[b].re*wre - x[b].im*wim;
	    x[b].im = x[a].im - (xbim = x[b].re*wim + x[b].im*wre);
	    x[b].re = x[a].re - xbre;

	    x[a].re += xbre;
	    x[a].im += xbim;
	 }
      }
   }
}

void FFT2(cplx far x[], int N)
/* Klasicni FFT radiks-2 algoritam sa preuredjenjem u vremenu (DIT). */
/* sinusi/kosinusi racunaju se rekurzivno: brze ali unosi malu gresku. */
{
   int visina_leptira, visina_leptira_2, br_delova, deo, leptir, a, b, s;
   float fi, c = -2*M_PI/N;
   float wre, wim;
   float wre0, wim0;
   float dwre, dwim;
   float xbre, xbim;

   for (visina_leptira = 1; visina_leptira < N; visina_leptira <<= 1) {
      br_delova = N / (visina_leptira_2 = visina_leptira << 1);
      wre = 1.0;
      wim = 0.0;
      fi = c * br_delova;
      dwre = cos(fi);
      dwim = sin(fi);
      for (leptir = 0, s = 0; leptir < visina_leptira; leptir++, s += br_delova) {
	 for (a = leptir; a < N; a += visina_leptira_2) {
	    b = a + visina_leptira;

	    xbre = x[b].re*wre - x[b].im*wim;
	    x[b].im = x[a].im - (xbim = x[b].re*wim + x[b].im*wre);
	    x[b].re = x[a].re - xbre;

	    x[a].re += xbre;
	    x[a].im += xbim;
	 }
	 /* promena ugla w = w * dw */
	 wre0 = wre;
	 wim0 = wim;

	 wre = wre0*dwre - wim0*dwim;
	 wim = wre0*dwim + wim0*dwre;
      }
   }
}

void FFT3(cplx far x[], int N, float far cosfi[], float far sinfi[])
/* Klasicni FFT radiks-2 algoritam sa preuredjenjem u vremenu (DIT). */
/* Kosinusi i sinusi uzimaju se iz tabele. */
{
   int visina_leptira, visina_leptira_2, br_delova, deo, leptir, a, b, s;
   float wre, wim;
   float xbre, xbim;

   for (visina_leptira = 1; visina_leptira < N; visina_leptira <<= 1) {
      br_delova = N / (visina_leptira_2 = visina_leptira << 1);
      for (leptir = 0, s = 0; leptir < visina_leptira; leptir++, s += br_delova) {
	 wre = cosfi[s];
	 wim = sinfi[s];
	 for (a = leptir; a < N; a += visina_leptira_2) {
	    b = a + visina_leptira;

	    xbre = x[b].re*wre - x[b].im*wim;
	    x[b].im = x[a].im - (xbim = x[b].re*wim + x[b].im*wre);
	    x[b].re = x[a].re - xbre;

	    x[a].re += xbre;
	    x[a].im += xbim;
	 }
      }
   }
}

void FFT4(cplx far x[], int N, float far cosfi[], float far sinfi[])
/* Klasicni FFT radiks-2 algoritam sa preuredjenjem u vremenu (DIT). */
/* Kosinusi i sinusi uzimaju se iz tabele. Koristi pointere. */
{
   int visina_leptira, visina_leptira_2, br_delova, deo, leptir, a;
   float wre, wim;
   float xbre, xbim;
   cplx far *pxa;
   cplx far *pxb;
   float far *pcos;
   float far *psin;

   for (visina_leptira = 1; visina_leptira < N; visina_leptira <<= 1) {
      br_delova = N / (visina_leptira_2 = visina_leptira << 1);
      pcos = cosfi;
      psin = sinfi;
      for (leptir = 0; leptir < visina_leptira; leptir++) {
	 wre = *pcos;
	 wim = *psin;
	 pcos += br_delova;
	 psin += br_delova;
	 for (a = leptir; a < N; a += visina_leptira_2) {
	    pxa = x + a;
	    pxb = pxa + visina_leptira;

	    xbre = pxb->re*wre - pxb->im*wim;
	    pxb->im = pxa->im - (xbim = pxb->re*wim + pxb->im*wre);
	    pxb->re = pxa->re - xbre;

	    pxa->re += xbre;
	    pxa->im += xbim;
	 }
      }
   }
}

double get_f0(cplx far f[], int size, double sync_level, int h_end)
/* Braca digitalnu frekvenciju osnovnog harmonika. */
/* Prvo pronadje maksimalnu frekv. komponentu koja je veca od sync_level */
/* To je PRIBLIZNO osnovna frekvencija signala. */
/* Zatim ispita maksimume u okolinama frekvencijskih komponenti koje su */
/* multipli osnovne frekvencije. Na kraju se izracuna prosek odstojanja */
/* frekv. maksimuma (harmonici) i to je osnovna frekvencija signala */
/* Prilikom racunanja proseka uzima se tezinska srednja vrednost, tj. */
/* frekvencijska komponenta sa vecom amplitudom vise utice na prosek, dok */
/* oni maksimumi koji su manji od sync_level uopste ne uticu na prosek. */
/* Ako je h_end >= 1 razmatra se maksimalno h_end harmonika, u protivnom */
/* svi postojeci (h_end <= 0). */
/* Ovo vazi akoje sync_level > 0.0. Ako je = 0.0 onda se vraca frekvencija */
/* onog harmonika koji ima najvecu amplitudu */
{
   int k_width = 8;
   int size2 = size/2;
   int k, k_end, k_max;	/* digitalna frekvencija */
   double a, a_max;     /* amplituda */
   double a_sum, b_sum;	/* za racunanje srednje vrednosti a_sum/b_sum */
   double f0_dig;	/* trazena osnovna frekvencija */
   int h;		/* harmonik */

   if (sync_level > 0.0) {	/* trazi prvi maksimum veci od sync_level */
      for (k = 2; k < size2; k++) {
	 a = sqrt(f[k].re*f[k].re + f[k].im*f[k].im);
	 if (a >= sync_level)
	    break;
      }
      a_max = 0.0;
      k_end = k + k_width;
      k_max = k;
      while (k < k_end && k < size2) {
	 a = sqrt(f[k].re*f[k].re + f[k].im*f[k].im);
	 if (a_max < a) {
	    a_max = a;
	    k_max = k;
	 }
	 k++;
      }
   }
   else {			/* trazi najveci maksimum */
      a_max = 0.0;
      k_max = 2;
      for (k = 2; k < size2; k++) {
	 a = sqrt(f[k].re*f[k].re + f[k].im*f[k].im);
	 if (a_max < a) {
	    a_max = a;
	    k_max = k;
	 }
      }

      return k_max;
   }

   if (a_max == 0.0)
      return 0.0;

   k_width = k_max / 2;

   f0_dig = k_max;
   a_sum = a_max * f0_dig;
   b_sum = a_max;

   if (h_end <= 0)
      h_end = size2 / k_max;
   for (h = 2; h < h_end; h++) {   /* Ostali harmonici (ako postoje) */

      /* Trazi maksimum u u opsegu k = h*f0_dig +/- k_width/2 */
      k = h*f0_dig - k_width/2;
      k_end += k_width;
      a_max = 0;
      do {
	 a = sqrt(f[k].re*f[k].re + f[k].im*f[k].im);
	 if (a_max < a) {
	    a_max = a;
	    k_max = k;
	 }
	 k++;
      }  while (k < k_end && k < size2);

      if (a_max >= sync_level) {	/* Ako je maksimum zadovoljavajuci */
	 f0_dig = (double) k_max / h;
	 a_sum += a_max * f0_dig;
	 b_sum += a_max;

	 f0_dig = a_sum / b_sum;	/* Nova srednja vrednost f0_dig */
      }
   }

   return f0_dig;
}

cplx get_fourier(cplx far x[], int N, double omega)
/* Na osnovu DFT sekvence x[] daje furijeovu transformaciju dig. signala. */
{
   int k;
   int N1 = N - 1;
   double c1 = M_PI/N;
   double fi;
   double a;
   double sum_re = 0.0, sum_im = 0.0;
   double p_re, p_im;
   cplx z;

   omega /= 2;
   for (k = 0; k < N; k++) {
      fi = omega - c1 * k;

      if (fi != 0.0)
	 a = sin(fi*N) / sin(fi);
      else
	 a = 1.0;

      fi *= N1;

      p_re = a*cos(fi);
      p_im = -a*sin(fi);

      sum_re += x[k].re * p_re - x[k].im * p_im;
      sum_im += x[k].re * p_im + x[k].im * p_re;
   }

   z.re = sum_re / N;
   z.im = sum_im / N;

   return z;
}

